package cc.fascinated.bat.common;

import cc.fascinated.bat.Consts;
import lombok.AllArgsConstructor;
import lombok.NonNull;
import net.dv8tion.jda.api.EmbedBuilder;
import net.dv8tion.jda.api.entities.Message;
import net.dv8tion.jda.api.entities.MessageEmbed;
import net.dv8tion.jda.api.interactions.InteractionHook;
import net.dv8tion.jda.api.interactions.callbacks.IReplyCallback;
import net.dv8tion.jda.api.interactions.components.LayoutComponent;
import net.dv8tion.jda.api.requests.RestAction;
import net.dv8tion.jda.api.requests.restaction.WebhookMessageCreateAction;
import net.dv8tion.jda.api.requests.restaction.interactions.ReplyCallbackAction;

import java.awt.*;

/**
 * A simple utility class to make
 * responding to interactions easier.
 *
 * @author Braydon
 */
@AllArgsConstructor
public final class InteractionResponder {
    /**
     * The interaction event to respond to.
     */
    @NonNull private final IReplyCallback event;

    /**
     * Respond to the interaction with a generic embed.
     *
     * @param content the content to respond with
     * @param args    the optional arguments to format the content with
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replyGenericEmbed(@NonNull String content, Object... args) {
        return replyGenericEmbed(content, args, (LayoutComponent[]) null);
    }

    /**
     * Respond to the interaction with a generic embed.
     *
     * @param content the content to respond with
     * @param args    the optional arguments to format the content with
     * @param components the optional components to add to the message
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replyGenericEmbed(@NonNull String content, Object[] args, LayoutComponent... components) {
        return replyGenericEmbed(false, content, args, components);
    }

    /**
     * Respond to the interaction with a generic embed.
     *
     * @param ephemeral whether the message should be ephemeral
     * @param content   the content to respond with
     * @param args      the optional arguments to format the content with
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replyGenericEmbed(boolean ephemeral, @NonNull String content, Object... args) {
        return replyGenericEmbed(ephemeral, content, args, (LayoutComponent[]) null);
    }

    /**
     * Respond to the interaction with a generic embed.
     *
     * @param ephemeral  whether the message should be ephemeral
     * @param content    the content to respond with
     * @param components the optional components to add to the message
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replyGenericEmbed(boolean ephemeral, @NonNull String content, LayoutComponent... components) {
        return replyGenericEmbed(ephemeral, content, null, components);
    }

    /**
     * Respond to the interaction with a generic embed.
     *
     * @param ephemeral  whether the message should be ephemeral
     * @param content    the content to respond with
     * @param args       the optional arguments to format the content with
     * @param components the optional components to add to the message
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replyGenericEmbed(boolean ephemeral, @NonNull String content, Object[] args, LayoutComponent... components) {
        content = format(content, args);
        return reply(ephemeral, buildGenericEmbed(content, args), components);
    }

    /**
     * Build a generic embed.
     *
     * @param content the content to respond with
     * @param args    the optional arguments to format the content with
     * @return the embed
     */
    @NonNull
    public MessageEmbed buildGenericEmbed(@NonNull String content, Object... args) {
        content = format(content, args);
        return new EmbedBuilder()
                .setColor(Colors.DEFAULT)
                .setDescription(format(content, args))
                .build();
    }

    /**
     * Respond to the interaction with a success embed.
     *
     * @param content the content to respond with
     * @param args    the optional arguments to format the content with
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replySuccess(@NonNull String content, Object... args) {
        return replySuccess(false, content, args);
    }

    /**
     * Respond to the interaction with a success embed.
     *
     * @param content    the content to respond with
     * @param args       the optional arguments to format the content with
     * @param components the optional components to add to the message
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replySuccess(@NonNull String content, Object[] args, LayoutComponent... components) {
        return replySuccess(false, content, args, components);
    }

    /**
     * Respond to the interaction with a success embed.
     *
     * @param ephemeral whether the message should be ephemeral
     * @param content   the content to respond with
     * @param args      the optional arguments to format the content with
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replySuccess(boolean ephemeral, @NonNull String content, Object... args) {
        return replySuccess(ephemeral, content, args, (LayoutComponent[]) null);
    }

    /**
     * Respond to the interaction with a success embed.
     *
     * @param ephemeral  whether the message should be ephemeral
     * @param content    the content to respond with
     * @param args       the optional arguments to format the content with
     * @param components the optional components to add to the message
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replySuccess(boolean ephemeral, @NonNull String content, Object[] args, LayoutComponent... components) {
        content = format(content, args);
        return reply(ephemeral, new EmbedBuilder()
                .setColor(Color.GREEN)
                .setDescription(BatEmoji.SUCCESS + " " + content)
                .build(), components);
    }

    /**
     * Respond to the interaction with an error embed.
     *
     * @param error the error to respond with
     * @param args  the optional arguments to format the error with
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replyError(@NonNull String error, Object... args) {
        return replyError(false, error, args);
    }

    /**
     * Respond to the interaction with an error embed.
     *
     * @param ephemeral whether the message should be ephemeral
     * @param error     the error to respond with
     * @param args      the optional arguments to format the error with
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> replyError(boolean ephemeral, @NonNull String error, Object... args) {
        error = format(error, args);
        return reply(ephemeral, new EmbedBuilder()
                .setColor(Color.RED)
                .setDescription(BatEmoji.CROSS + " " + error)
                .build());
    }

    /**
     * Respond to the interaction with an embed.
     *
     * @param embed the embed to respond with
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> reply(@NonNull MessageEmbed embed) {
        return reply(false, embed);
    }

    /**
     * Respond to the interaction with an embed.
     *
     * @param embed      the embed to respond with
     * @param components the optional components to add to the message
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> reply(@NonNull MessageEmbed embed, LayoutComponent... components) {
        return reply(false, embed, components);
    }

    /**
     * Respond to the interaction with an embed.
     *
     * @param ephemeral whether the message should be ephemeral
     * @param embed     the embed to respond with
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> reply(boolean ephemeral, @NonNull MessageEmbed embed) {
        return reply(ephemeral, embed, (LayoutComponent[]) null);
    }

    /**
     * Respond to the interaction with an embed.
     *
     * @param ephemeral  whether the message should be ephemeral
     * @param embed      the embed to respond with
     * @param components the optional components to add to the message
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> reply(boolean ephemeral, @NonNull MessageEmbed embed, LayoutComponent... components) {
        boolean hasComponents = components != null && (components.length > 0);
        if (event.isAcknowledged()) {
            WebhookMessageCreateAction<Message> action = event.getHook()
                    .setEphemeral(ephemeral)
                    .sendMessageEmbeds(embed);
            return hasComponents ? action.setComponents(components) : action;
        }
        ReplyCallbackAction action = event.replyEmbeds(embed).setEphemeral(ephemeral);
        return (hasComponents ? action.setComponents(components) : action).flatMap(InteractionHook::retrieveOriginal);
    }

    /**
     * Respond to the interaction with a message.
     *
     * @param content the message to respond with
     * @param args    the optional arguments to format the message with
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> reply(@NonNull String content, Object... args) {
        return reply(false, content, args);
    }

    /**
     * Respond to the interaction with a message.
     *
     * @param ephemeral whether the message should be ephemeral
     * @param content   the message to respond with
     * @param args      the optional arguments to format the message with
     * @return the reply action
     */
    @NonNull
    public RestAction<Message> reply(boolean ephemeral, @NonNull String content, Object... args) {
        content = format(content, args);
        if (event.isAcknowledged()) {
            return event.getHook().setEphemeral(ephemeral).sendMessage(content);
        }
        return event.reply(content).setEphemeral(ephemeral).flatMap(InteractionHook::retrieveOriginal);
    }

    /**
     * Format the given content with the given arguments.
     *
     * @param content the content to format
     * @param args    the arguments to format with
     * @return the formatted content
     */
    @NonNull
    private String format(@NonNull String content, Object... args) {
        return args != null && (args.length > 0) ? content.formatted(args) : content;
    }
}