package cc.fascinated.bat.common;

import lombok.NonNull;
import lombok.experimental.UtilityClass;
import net.dv8tion.jda.api.JDA;

import java.util.Arrays;
import java.util.Objects;
import java.util.stream.Collectors;

@UtilityClass
public final class MiscUtils {
    /**
     * Get the invite url for the bot.
     *
     * @param jda the JDA instance
     * @return the invite url
     */
    @NonNull
    public static String getInviteUrl(@NonNull JDA jda) {
        return "https://discord.com/oauth2/authorize?client_id=" + jda.getSelfUser().getId();
    }

    /**
     * Convert an array of strings to a single string.
     *
     * @param lines the lines to convert
     * @return the string
     */
    @NonNull
    public static String arrayToString(String... lines) {
        return Arrays.stream(lines)
                .filter(Objects::nonNull)
                .collect(Collectors.joining("\n"));
    }

    /**
     * Capitalize the first letter of
     * every word in the given string.
     *
     * @param input the input to capitalize
     * @return the capitalized string
     */
    @NonNull
    public static String capitalize(@NonNull String input) {
        input = input.toLowerCase().replace('_', ' ');
        StringBuilder result = new StringBuilder();
        boolean capitalizeNext = true;
        for (char character : input.toCharArray()) {
            if (Character.isSpaceChar(character)) {
                capitalizeNext = true;
                result.append(character);
            } else if (capitalizeNext) {
                result.append(Character.toUpperCase(character));
                capitalizeNext = false;
            } else {
                result.append(character);
            }
        }
        return result.toString();
    }

    /**
     * Get the enum constant with the given
     * name from the given enum class.
     *
     * @param enumClass    the enum class
     * @param enumName     the enum name
     * @param <T>          the enum type
     * @return the enum constant, null if not found
     */
    public static <T extends Enum<T>> T getEnum(@NonNull Class<T> enumClass, String enumName) {
        return getEnum(enumClass, enumName, null);
    }

    /**
     * Get the enum constant with the given
     * name from the given enum class.
     *
     * @param enumClass    the enum class
     * @param enumName     the enum name
     * @param defaultValue the default value
     * @param <T>          the enum type
     * @return the enum constant, default value if not found
     */
    public static <T extends Enum<T>> T getEnum(@NonNull Class<T> enumClass, String enumName, T defaultValue) {
        if (enumName == null) {
            return defaultValue;
        }
        try {
            return Enum.valueOf(enumClass, enumName);
        } catch (IllegalArgumentException ex) {
            return defaultValue;
        }
    }
}
