package cc.fascinated.bat.service;

import cc.fascinated.bat.Consts;
import cc.fascinated.bat.command.BatCommand;
import cc.fascinated.bat.command.BatSubCommand;
import cc.fascinated.bat.command.CommandCategory;
import cc.fascinated.bat.command.IBatCommand;
import cc.fascinated.bat.command.annotation.BotOwner;
import cc.fascinated.bat.command.annotation.Thinking;
import cc.fascinated.bat.command.command.HelpCommand;
import cc.fascinated.bat.common.InteractionResponder;
import cc.fascinated.bat.common.model.BatGuild;
import jakarta.annotation.Nonnull;
import jakarta.annotation.PostConstruct;
import lombok.Getter;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import net.dv8tion.jda.api.Permission;
import net.dv8tion.jda.api.entities.Guild;
import net.dv8tion.jda.api.entities.Member;
import net.dv8tion.jda.api.events.interaction.command.CommandAutoCompleteInteractionEvent;
import net.dv8tion.jda.api.events.interaction.command.SlashCommandInteractionEvent;
import net.dv8tion.jda.api.hooks.ListenerAdapter;
import net.dv8tion.jda.api.interactions.commands.Command;
import net.dv8tion.jda.api.interactions.commands.CommandInteractionPayload;
import net.dv8tion.jda.api.interactions.commands.OptionMapping;
import net.dv8tion.jda.api.interactions.commands.build.SlashCommandData;
import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.info.BuildProperties;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

/**
 * @author Braydon
 */
@Service @Log4j2(topic = "Commands")
public final class CommandService extends ListenerAdapter {
    /**
     * The guild service to get and create guilds.
     */
    public final GuildService guildService;

    /**
     * The build properties for this app, null if not available.
     */
    private final BuildProperties buildProperties;

    /**
     * The currently registered commands.
     */
    @Getter private final List<BatCommand> registry = Collections.synchronizedList(new ArrayList<>());

    @Autowired
    public CommandService(@NonNull GuildService guildService, BuildProperties buildProperties) {
        this.guildService = guildService;
        this.buildProperties = buildProperties;
        DiscordService.JDA.addEventListener(this);
    }

    @PostConstruct
    public void onInitialize() {
        // Command Registry
        registerCommand(new HelpCommand(this, buildProperties));
        log.info("Registered {} commands", registry.size());

        // Send the commands to the guilds
        log.info("Sending {} commands to guilds...", registry.size());
        List<SlashCommandData> commandData = registry.stream()
                .map(BatCommand::getCommandData)
                .toList();

        // the amount of indents here should be illegal
        for (Guild guild : DiscordService.JDA.getGuilds()) {
            log.info("Queueing {} commands for guild {}", commandData.size(), guild.getName());
            guild.updateCommands().addCommands(commandData).queue(registry -> {
                // Once sent to Discord, set the snowflake of the commands in our registry
                for (Command registeredCommand : registry) {
                    BatCommand rufusCommand = Objects.requireNonNull(getCommand(registeredCommand.getName()));
                    rufusCommand.setSnowflake(registeredCommand.getIdLong());
                    log.info("Retrieved snowflake {} for command /{}", rufusCommand.getSnowflake(), rufusCommand.getName());

                    // Register the sub-command groups
                    for (Command.SubcommandGroup subCommandGroup : registeredCommand.getSubcommandGroups()) {
                        for (Command.Subcommand subCommand : subCommandGroup.getSubcommands()) {
                            String fullName = rufusCommand.getName() + " " + subCommandGroup.getName() + " " + subCommand.getName();
                            BatSubCommand rufusSubCommand = Objects.requireNonNull(rufusCommand.getSubCommand(subCommandGroup.getName(), subCommand.getName()));
                            rufusSubCommand.setSnowflake(subCommand.getIdLong());
                            rufusSubCommand.setFullName(fullName);
                            log.info("Retrieved snowflake {} for sub-command-group /{}", subCommand.getIdLong(), fullName);
                        }
                    }

                    // Register the sub-commands
                    for (Command.Subcommand subCommand : registeredCommand.getSubcommands()) {
                        String fullName = rufusCommand.getName() + " " + subCommand.getName();
                        BatSubCommand rufusSubCommand = Objects.requireNonNull(rufusCommand.getSubCommand(null, subCommand.getName()));
                        rufusSubCommand.setSnowflake(subCommand.getIdLong());
                        rufusSubCommand.setFullName(fullName);
                        log.info("Retrieved snowflake {} for sub-command /{}", subCommand.getIdLong(), fullName);
                    }
                }
            });
        }
        log.info("Done, commands have been sent to {} guilds", DiscordService.JDA.getGuilds().size());
    }

    @Override
    public void onSlashCommandInteraction(@NotNull SlashCommandInteractionEvent event) {
        Member member = event.getMember();
        Guild discordGuild = event.getGuild();
        assert member != null && discordGuild != null;
        BatGuild guild = guildService.getGuild(discordGuild.getId());

        // Execute the command if registered
        for (BatCommand parent : registry) {
            if (!parent.getName().equals(event.getName())) {
                continue;
            }
            // Log the executed command
            logUsage(discordGuild, guild, member, parent, event.getSubcommandGroup(), event.getSubcommandName(), event);

            // If @Thinking is present, tell discord we received the command, send a thinking... message to the user
            if (parent.getClass().isAnnotationPresent(Thinking.class)) {
                event.deferReply().queue();
            }
            InteractionResponder responder = new InteractionResponder(event);
            IBatCommand executor = getExecutor(parent, event);

            // Initial checks before running the command
            if (executor.getClass().isAnnotationPresent(BotOwner.class) && !member.getId().equals(Consts.BOT_OWNER)) {
                responder.replyError("This command is restricted to the bot owner.").queue();
                return;
            }

            // Execute the command
            executor.execute(guild, discordGuild, member, responder, event);
            break;
        }
    }

    @Override
    public void onCommandAutoCompleteInteraction(@NotNull CommandAutoCompleteInteractionEvent event) {
        Member member = event.getMember();
        Guild discordGuild = event.getGuild();
        assert member != null && discordGuild != null;
        BatGuild guild = guildService.getGuild(discordGuild.getId());

        // Auto complete the command if registered
        IBatCommand executor = getExecutor(event);
        if (executor == null) {
            return;
        }
        List<Command.Choice> choices = executor.autoComplete(guild, discordGuild, member, event);
        if (choices.isEmpty()) {
            return;
        }
        event.replyChoices(choices.stream()
                .filter(choice -> choice.getName().startsWith(event.getFocusedOption().getValue()))
                .collect(Collectors.toList())).queue();
    }

    /**
     * Register a command.
     *
     * @param command the command
     */
    private void registerCommand(@NonNull BatCommand command) {
        registry.add(command);
        log.info("Registered command: /{}", command.getCommandData().getName());
    }

    /**
     * Get the commands in a category
     * the given member has access to.
     *
     * @param member the member to get the commands for
     * @param category the category to get the commands for
     * @return the commands the member has
     */
    @NonNull
    public List<IBatCommand> getCommands(@Nonnull Member member, @NonNull CommandCategory category) {
        // Collect all the commands in this category
        List<IBatCommand> categoryCommands = registry.stream()
                .filter(command -> {
                    // Not the right category
                    if (command.getCategory() != category) {
                        return false;
                    }
                    // Return the commands the user has access to
                    Long rawPermissions = command.getCommandData().getDefaultPermissions().getPermissionsRaw();
                    return rawPermissions == null || (member.getPermissions().containsAll(Permission.getPermissions(rawPermissions)));
                })
                .collect(Collectors.toList());

        // Add all the sub-commands from the parents
        List<IBatCommand> subCommands = categoryCommands.stream()
                .flatMap(parent -> ((BatCommand) parent).getSubCommands().stream())
                .collect(Collectors.toList());
        categoryCommands.addAll(subCommands);

        // Finally, remove any parents that have sub-commands as they can't be executed directly
        categoryCommands.removeIf(command -> command instanceof BatCommand parentCommand && (!parentCommand.getSubCommands().isEmpty()));
        return categoryCommands;
    }

    /**
     * Get a command by name.
     *
     * @param name the command name
     * @return the command, null if none
     */
    public BatCommand getCommand(@NonNull String name) {
        for (BatCommand command : registry) {
            if (command.getName().equalsIgnoreCase(name)) {
                return command;
            }
        }
        return null;
    }

    /**
     * Log the usage of a command.
     *
     * @param discordGuild    the Discord guild the command was executed in
     * @param guild           the ticket guild the command was executed in
     * @param member          the member that executed the command
     * @param parent          the parent command
     * @param subCommandGroup the sub-command group, if any
     * @param subCommandName  the sub-command name, if any
     * @param event           the event that triggered the command
     */
    private void logUsage(@NonNull Guild discordGuild, @NonNull BatGuild guild, @NonNull Member member, @NonNull BatCommand parent,
                          String subCommandGroup, String subCommandName, @NonNull SlashCommandInteractionEvent event) {
        StringBuilder fullCommandName = new StringBuilder(parent.getName());
        if (subCommandGroup != null) {
            fullCommandName.append(" ").append(subCommandGroup);
        }
        if (subCommandName != null) {
            fullCommandName.append(" ").append(subCommandName);
        }
        for (OptionMapping option : event.getOptions()) {
            fullCommandName.append(" ").append(option.getName()).append(":").append(option.getAsString());
        }
        log.info("User {} (UID: {}) executed '/{}' in guild {} (GID: {})",
                member.getUser().getName(), member.getId(), fullCommandName.toString(), discordGuild.getName(), guild.getId()
        );
    }

    /**
     * Get the command executor for the given event.
     *
     * @param event the event
     * @return the command executor, null if none
     */
    private IBatCommand getExecutor(@NonNull CommandInteractionPayload event) {
        for (BatCommand parent : registry) {
            if (!parent.getName().equals(event.getName())) {
                continue;
            }
            return getExecutor(parent, event);
        }
        return null;
    }

    /**
     * Get the command executor for the given command.
     *
     * @param parent the parent command
     * @param event the event
     * @return the command executor
     */
    @Nonnull
    private IBatCommand getExecutor(BatCommand parent, @NonNull CommandInteractionPayload event) {
        String subCommandGroup = event.getSubcommandGroup();
        String subCommandName = event.getSubcommandName();
        IBatCommand executor = null;

        // Get the command executor
        if (subCommandName != null) { // Try to find the sub-command to execute
            executor = parent.getSubCommand(subCommandGroup, subCommandName);
        }
        if (executor == null) { // Fallback to the parent
            executor = parent;
        }
        return executor;
    }
}